﻿using System;
using System.IO.Ports;

// PL2303_Prolific_DriverInstaller_v1.5.0.exe für Win8

namespace WS2812_USB_UART
{

    internal enum USB_Device
    {
        S560_PL2303HXA,     // 560 kBit PL2303HXA
        S560_PL2303HXB,     // 560 kBit PL2303HXB
        S560_PL2303TA,      // 560 kBit PL2303TA

        S680_PL2303HXA,     // 680 kBit PL2303HXA
        S680_PL2303HXB,     // 680 kBit PL2303HXB
        S680_PL2303TA,      // 680 kBit PL2303TA

        S780_PL2303HXA,     // 780 kBit PL2303HXA (Krumme Baudrate)
        S780_PL2303HXB,     // 780 kBit PL2303HXB (Krumme Baudrate)

        S840_PL2303HXA,     // 840 kBit PL2303HXA
        S840_PL2303HXB,     // 840 kBit PL2303HXA
        S840_PL2303TA,      // 840 kBit PL2303TA

        //FT232RL,          // Geht nicht bei 3MBaud, Verschluckt sporadisch Daten (Win7, XP gleiches Verhalten)
        //CP2102,           // Geht nicht Stoppbit zu lang / Baudrate zu gering 2MBaut
    }

    
    internal class WS_UART
    {

        #region Puplic Fields

        public readonly USB_Device Device;
        public readonly byte[] LutRgbUart;
        public readonly int LedCount;
        public readonly byte[] RGB_Array;
        public readonly byte[] UartArray;

        #endregion
        

        #region Constructor

        public WS_UART(USB_Device device) : this(device, 0) { }
        public WS_UART(USB_Device device, int ledCount)
        {
            Device = device;

            LutRgbUart = new byte[LutRgbUartSize];

            InitializeLUT();

            
            if (ledCount == 0)
                return;

            byte[] rgbArray = null;
            byte[] uartArray = null;

            CreateArrays(ref rgbArray, ref uartArray, ledCount);

            LedCount = ledCount;
            RGB_Array = rgbArray;
            UartArray = uartArray;

            RgbArrayToWsUartArray();
        }

        #endregion


        #region LUT RGB UART

        public const int LutRgbUartSize = 256 * 4;

        //      S0123 456ss   7-Bit
        //00    10000 10000   0x10
        //01    10000 11100   0x70
        //10    11100 10000   0x13
        //11    11100 11100   0x73
        // 
        // PL2023HXA / PL2023HXB / PL2023TA
        //
        private const byte P1_LutBit00 = 0x10 ^ 0xff;
        private const byte P1_LutBit01 = 0x70 ^ 0xff;
        private const byte P1_LutBit10 = 0x13 ^ 0xff;
        private const byte P1_LutBit11 = 0x73 ^ 0xff;
        private readonly byte[] P1_LutBit = { P1_LutBit00, P1_LutBit01, P1_LutBit10, P1_LutBit11 };

        //      S0123 4ss   5-Bit
        //00    10001 000   0x08
        //01    10001 100   0x18
        //10    11001 000   0x09
        //11    11001 100   0x19
        //
        // PL2023HXA / PL2023HXB / PL2023TA
        //
        private const byte P2_LutBit00 = 0x08 ^ 0xff;
        private const byte P2_LutBit01 = 0x18 ^ 0xff;
        private const byte P2_LutBit10 = 0x09 ^ 0xff;
        private const byte P2_LutBit11 = 0x19 ^ 0xff;
        private readonly byte[] P2_LutBit = { P2_LutBit00, P2_LutBit01, P2_LutBit10, P2_LutBit11 };

        //      S0123 456ss   7-Bit
        //00    11000 11000   0x31
        //01    11000 11100   0x71
        //10    11100 11000   0x33
        //11    11100 11100   0x73
        //
        // PL2023HXA / PL2023HXB
        //
        private const byte P3_LutBit00 = 0x31 ^ 0xff;
        private const byte P3_LutBit01 = 0x71 ^ 0xff;
        private const byte P3_LutBit10 = 0x33 ^ 0xff;
        private const byte P3_LutBit11 = 0x73 ^ 0xff;
        private readonly byte[] P3_LutBit = { P3_LutBit00, P3_LutBit01, P3_LutBit10, P3_LutBit11 };


        //      S0123 4567Pss   8-Bit + Parity
        //00    11000 0110000   0x61
        //01    11000 0111100   0xe1
        //10    11110 0110000   0x67
        //11    11110 0111100   0xe7
        //
        // PL2023HXA / PL2023HXB / PL2023TA
        //
        private const byte P4_LutBit00 = 0x61 ^ 0xff;
        private const byte P4_LutBit01 = 0xe1 ^ 0xff;
        private const byte P4_LutBit10 = 0x67 ^ 0xff;
        private const byte P4_LutBit11 = 0xe7 ^ 0xff;
        private readonly byte[] P4_LutBit = { P4_LutBit00, P4_LutBit01, P4_LutBit10, P4_LutBit11 };


        private void InitializeLUT()
        {

            byte[] LutBit = null;

            switch (Device)
            {
                case USB_Device.S560_PL2303HXA:
                case USB_Device.S560_PL2303HXB:
                case USB_Device.S560_PL2303TA:
                    LutBit = P1_LutBit;
                    break;

                case USB_Device.S680_PL2303HXA:
                case USB_Device.S680_PL2303HXB:
                case USB_Device.S680_PL2303TA:
                    LutBit = P2_LutBit;
                    break;
                
                case USB_Device.S780_PL2303HXA:
                case USB_Device.S780_PL2303HXB:
                    LutBit = P3_LutBit;
                    break;

                case USB_Device.S840_PL2303HXA:
                case USB_Device.S840_PL2303HXB:
                case USB_Device.S840_PL2303TA:
                    LutBit = P4_LutBit;
                    break;
            }
            
            int indexLut = 4;
            int j = 0;

            for (int i = 0; i < 256; i++)
            {
                for (j = 0; j < 8; j += 2)
                    LutRgbUart[--indexLut] = LutBit[(i >> j) & 0x03];
                indexLut += 8;
            }

        }

        #endregion


        #region Public Methods

        public static void CreateArrays(ref byte[] rgbArray, ref byte[] uartArray, int ledCount)
        {
            if (ledCount == 0)
                throw new Exception("ledCount == 0");

            rgbArray = new byte[ledCount * 3];
            uartArray = new byte[ledCount * 3 * 4];
        }


        public void SetupSerialPort(SerialPort serialPort)
        { SetupSerialPort(serialPort, Device, LedCount); }
        
        public static void SetupSerialPort(SerialPort serialPort, USB_Device dev, int ledCount)
        {
            serialPort.StopBits = StopBits.One;
            serialPort.WriteBufferSize = ledCount * 3 * 4;
            
            switch (dev)
            {
                case USB_Device.S560_PL2303HXA:
                case USB_Device.S560_PL2303HXB:
                case USB_Device.S560_PL2303TA:
                    serialPort.Parity = Parity.None;
                    serialPort.BaudRate = 3000000;
                    serialPort.DataBits = 7;
                    break;

                case USB_Device.S680_PL2303HXA:
                case USB_Device.S680_PL2303HXB:
                case USB_Device.S680_PL2303TA:
                    serialPort.Parity = Parity.None;
                    serialPort.BaudRate = 3000000;       
                    serialPort.DataBits = 5;
                    break;

                case USB_Device.S780_PL2303HXA:
                case USB_Device.S780_PL2303HXB:
                    serialPort.Parity = Parity.None;
                    serialPort.BaudRate = 4500000;
                    serialPort.DataBits = 7;
                    break;

                case USB_Device.S840_PL2303HXA:
                case USB_Device.S840_PL2303HXB:
                case USB_Device.S840_PL2303TA:
                    serialPort.Parity = Parity.Even;
                    serialPort.BaudRate = 6000000;
                    serialPort.DataBits = 8;
                    break;
            }
        }


        public void RgbArrayToWsUartArray()
        {
            RgbArrayToWsUartArray(RGB_Array, UartArray, LutRgbUart);
        }


        public static void RgbArrayToWsUartArray(byte[] rgbArray, byte[] uartArray, byte[] lutRgbUart)
        {
            if (rgbArray == null)
                throw new NullReferenceException("rgbArray");

            if (uartArray == null)
                throw new NullReferenceException("wsUartArray");

            if (lutRgbUart == null)
                throw new NullReferenceException("lutRgbUart");

            if (rgbArray.Length % 3 != 0)
                throw new Exception("rgbArray.Length is not valide");

            if (rgbArray.Length * 4 != uartArray.Length)
                throw new Exception("wsUartArray.Lengts is not valide");

            if (lutRgbUart.Length != LutRgbUartSize)
                throw new Exception("lutRgbUart.Length is not valide");


            int length = rgbArray.Length / 3;
            int iRgbA = 0;
            int iUartA = 0;
            int j, k1, k2, k3;

            for (int i = 0; i < length; i++)
            {
                k1 = rgbArray[iRgbA + 1] * 4;
                k2 = rgbArray[iRgbA + 0] * 4;
                k3 = rgbArray[iRgbA + 2] * 4;

                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k1 + j]; }
                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k2 + j]; }
                for (j = 0; j < 4; j++) { uartArray[iUartA++] = lutRgbUart[k3 + j]; }
                iRgbA += 3;
            }
        }


        #endregion
        
    }
}
